/**
 * charactron.c -- character drawing routines for SDL_reader, adapted for
 * unicode
 *
 * Tyler Montbriand, 2008
 * tsm@accesscomm.ca
 * Released under the GNU GPL
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef SCALE
#include "SDL_rotozoom.h"
#endif

#include "SDL.h"
#include "charactron.h"

#ifdef SCALE
void charactron_scale(charactron *c)
{
  SDL_Surface *n=zoomSurface(c->output,
	(double)SCALE_N/(double)SCALE_D,
	(double)SCALE_N/(double)SCALE_D,
	1);

  SDL_BlitSurface(n, NULL, c->scale, NULL);
  SDL_FreeSurface(n);
}
#endif


int charactron_set_color(charactron *c, int i, SDL_Color cl)
{
  if(i == CHARACTRON_BG)
    c->bg=SDL_MapRGB(c->output->format, cl.r, cl.g, cl.b);
  else if(i == CHARACTRON_FG)
    SDL_SetColors(c->font, &cl, 0, 1);
  else
    return(-1);

  return(0);  
}

void charactron_rect(const charactron *c, Uint32 letter, SDL_Rect *r)
{
	/**
	 * You'd think the X position in a grid of characters 8 wide
	 * would be a multiple of 8, wouldn't you?  WOULDN'T YOU?
	 *
	 * BUT IT'S NOT because SDL does weird things with 1bpp bitmaps.
	 *
	 * 1 horizontal position is 8 pixels.  For some reason.
	 */
	r->x=U_COL(letter);
	r->y=U_ROW(letter);
	r->w=8;
	r->h=CHAR_H;
}

void charactron_cursor_rect(const charactron *c, SDL_Rect *r)
{
	r->w=CHARACTRON_FONT_W(c);
	r->h=CHARACTRON_FONT_H(c);
	r->x=(r->w * c->x);
	r->y=(r->h * c->y);
}


void charactron_free(charactron *c)
{
  if(c != NULL)
  {
    if(c->font != NULL)
      SDL_FreeSurface(c->font);

    if((c->output != NULL)&&(c->output != SDL_GetVideoSurface()))
      SDL_FreeSurface(c->output);

    free(c);
  }
}

charactron *charactron_load(unifont *font, SDL_Surface *out, int autoclose)
{
  charactron *c=NULL;

  if(font==NULL)
      goto CHARACTRON_LOAD_ERR;

  if(out == NULL)
  {
    out=SDL_GetVideoSurface();
    if(out == NULL)
      goto CHARACTRON_LOAD_ERR;
  }

  c=(charactron *)malloc(sizeof(charactron));
  if(c==NULL)
    goto CHARACTRON_LOAD_ERR;

#ifdef SCALE
  c->scale=out;
  c->output=SDL_CreateRGBSurface(SDL_SWSURFACE,
    (out->w*SCALE_D)/SCALE_N, (out->h*SCALE_D)/SCALE_N, 
    out->format->BitsPerPixel,
    out->format->Rmask, out->format->Gmask,
    out->format->Bmask, out->format->Amask);
#else
  c->output=out;
#endif

  c->rawfont=font;

  c->font=SDL_CreateRGBSurfaceFrom(font->charset_l,
    U_IMG_W, U_IMG_H, U_IMG_BPP, U_IMG_PITCH, 0, 0, 0, 0);

  if(c->font == NULL)
    goto CHARACTRON_LOAD_ERR;

//  SDL_BlitSurface(c->font, NULL, c->output, NULL);
//  SDL_Flip(c->output);
//  SDL_Delay(5000);
//  abort();

  c->dirx=0;
  c->maxx=c->output->w / CHARACTRON_FONT_W(c);
  c->maxy=c->output->h / CHARACTRON_FONT_H(c);
  c->flag=CH_ENABLE;

  fprintf(stderr, "Display is (%d,%d)\n", c->maxx, c->maxy);

  c->bg=SDL_MapRGB(c->output->format,255,255,255);
  c->fg=SDL_MapRGB(c->output->format,  0,  0,  0);

  if(c->font->format->BitsPerPixel > 8)
  {
    SDL_SetError("Font must be a paletted bitmap");
    goto CHARACTRON_LOAD_ERR;
  }

  SDL_SetColorKey(c->font, SDL_SRCCOLORKEY,
    SDL_MapRGB(c->font->format, 0, 0, 0));

  {
    SDL_Color clr={0, 0, 0, 0};
    int index=SDL_MapRGB(c->font->format, 255, 255, 255);
    SDL_SetColors(c->font, &clr, index, 1);
  }


  charactron_formfeed(c);

  return(c);

CHARACTRON_LOAD_ERR:
  charactron_free(c);
  return(NULL);
}

int charactron_draw_letter(charactron *c, Uint32 letter)
{
  int w=1;
  SDL_Rect src,dest;

  if(!UNIFONT_ENCODED(c->rawfont, letter))
    letter=65533;
  else if(UNIFONT_WIDE(c->rawfont, letter))
    w++;

  if(c->xfree < w)
    return(-1);

  if(CHARACTRON_ENABLED(c))
  {
    charactron_cursor_rect(c, &dest);
    charactron_rect(c, letter, &src);

    SDL_BlitSurface(c->font,&src,c->output,&dest);

    if(w == 2)
    {
      dest.x+=CHARACTRON_X_OFF(c)*CHARACTRON_FONT_W(c);
      charactron_rect(c, letter^0x10000, &src);
      SDL_BlitSurface(c->font,&src,c->output,&dest);
    }
  }

  c->x += CHARACTRON_X_OFF(c)*w;
  c->xfree-=w;

  return(0);
}

#define SNAP_DRAW(C, Q, S, D) do {				\
	charactron_rect((C), (Q), &(S));			\
	SDL_BlitSurface((C)->font,&(S),(C)->output,&(D));	\
	(C)->x+=CHARACTRON_X_OFF(c);				\
	(D).x+=CHARACTRON_X_OFF(c)*CHARACTRON_FONT_W(c);	\
	} while(0)

// Note:  Will happily draw off end of screen, so use with caution
int charactron_draw_data(charactron *c, const Uint8 *mem,
	mempos_t *start, mempos_t end)
{
  int n=0, g;
  SDL_Rect src, dest;

  if(!CHARACTRON_ENABLED(c))
  {
    while((g=utf8_getchar(mem, start, end))>=0)
    {
      if(g < ' ') continue;

      if(!UNIFONT_ENCODED(c->rawfont, g))
        g=65533;

      n++;
      if(UNIFONT_WIDE(c->rawfont, g)) n++;
    }

    c->x+=CHARACTRON_X_OFF(c)*n;
  }
  else
  {
    charactron_cursor_rect(c, &dest);

    while((g=utf8_getchar(mem, start, end))>=0)
    {
      if(g < ' ') continue;

      SNAP_DRAW(c, g, src, dest);
      n++;

      if(UNIFONT_WIDE(c->rawfont, g))
      {
        SNAP_DRAW(c, g^0x10000, src, dest);
        n++;
      }
    }
  }

  c->xfree-=n;
  return(n);
}

int charactron_linefeed(charactron *c)
{
  if(c->y >= (c->maxy - 1))
    return(-1);

  c->y++;
//  c->x=(c->maxx + c->dirx)%(c->maxx);
//  c->xfree=c->maxx;
  CHARACTRON_X_RESET(c);
  return(0);
}

int charactron_formfeed(charactron *c)
{
//  Uint32 clr=SDL_MapRGB(c->output->format, 255, 0, 255);
  if(CHARACTRON_ENABLED(c))
    SDL_FillRect(c->output, NULL, c->bg);
//    SDL_FillRect(c->output,NULL, clr);

//  c->x=(c->maxx + c->dirx)%(c->maxx);
//  c->xfree=c->maxx;
  CHARACTRON_X_RESET(c);
  c->y=0;
  return(0);
}
