/**
 * Creates unifont.bin from unifont.hex.  unifont.bin is a flat file with the
 * following layout:
 *
 * 0x000000-0x001fff:	Bitmask of font widths.  0=narrow, 1=wide
 * 0x002000-0x003fff:	Bitmask of character presence.  0=unencoded, 1=encoded
 * 0x004000-0x103fff:	8-wide bitmasks for left half(or only half, if narrow)
 * 0x104000-0x203fff:	8-wide bitmasks for right half(blank, if narrow)
 *
 * Splitting it into low and high halves like this may mean space savings when
 * stored as a sparse file.
 * See unifont.h for exact arrangement of the bitmasks.
 *
 * Tyler Montbriand, 2008
 * tsm@accesscomm.ca
 * Released under the GNU GPL
 */
 
#include <stdio.h>
#include <string.h>
#include "unifont.h"
#include "SDL_mmap.h"
#include "SDL_endian.h"

#ifdef WIN32
#define FONT_INPUT	"../unifont-dvdeug-1.0/unifont.hex"
#define FONT_OUTPUT "../unifont.bin"
#else
#define FONT_INPUT "unifont-dvdeug-1.0/unifont.hex"
#define FONT_OUTPUT "unifont.bin"
#endif

int main(int argc, char *argv[])
{
	int count=0;
	char buf[128];

	struct SDL_mmap *mm=NULL;
	SDL_mspan *ms=NULL;
	FILE *fp=fopen(FONT_INPUT, "r");
	unifont *uf;

	if(fp == NULL)
	{
		fprintf(stderr, "Couldn't open unifont.hex\n");
		return(1);
	}

	mm=SDL_mmap_open(FONT_OUTPUT, SDL_MMAP_RD|SDL_MMAP_WR, 
		SDL_MADVISE_SEQUENTIAL);
	if(mm == NULL)
	{
		fprintf(stderr, "Couldn't map: %s\n", SDL_GetError());
		return(1);
	}

	SDL_mmap_truncate(mm, (mmap_t)0);
	SDL_mmap_truncate(mm, sizeof(unifont));
//	SDL_mmap_close(mm);

	ms=SDL_mmap_mapfile(FONT_OUTPUT, SDL_MMAP_RD|SDL_MMAP_WR,
		SDL_MADVISE_SEQUENTIAL);
	if(ms == NULL)
	{
		fprintf(stderr, "Couldn't map %s: %s\n",
			FONT_OUTPUT, SDL_GetError());
		exit(1);
	}

	uf=(unifont *)(ms->mem);

	while(!feof(fp))
	{
		int glyph, wide=0, pos;
		char *tmp;

		buf[0]='\0';
		if(fgets(buf, 127, fp) == NULL) continue;
		buf[127]='\0';

		if(sscanf(buf, "%x:", &glyph) != 1)
			continue;
		if(glyph > 65535)
		{
			fprintf(stderr, "glyph out of range\n");
			continue;
		}

		if(strlen(buf) > 45)
		{
			uf->glyphsize[glyph/8]|=BIT(glyph&7);
			wide=1;
		}

		uf->encoded[glyph/8]|=BIT(glyph&7);

		printf("Glyph: %04x (wide=%d)\r", glyph, wide);

		tmp=buf+5;

		if(wide)
		{
			int word;
			unsigned int val, val2;

			for(word=0; word<16; word++)
			{
				sscanf(tmp+(word*4), "%02x%02x", &val, &val2);

				L_BYTE(uf, glyph, word)=val^0xff;
				H_BYTE(uf, glyph, word)=val2^0xff;
			}
		}
		else
		{
			int word, val;

			for(word=0; word<16; word++)
			{
				sscanf(tmp+(word*2), "%02x", &val);
				L_BYTE(uf, glyph, word)=val^0xff;
			}
		}
	}

	fclose(fp);
	return(0);
}
