/**
 * text_parser.c -- text parsing routines for SDL_reader, nearly completely
 * rewritten for unicode support.
 *
 * Tyler Montbriand, 2008
 * tsm@accesscomm.ca
 * Released under the GNU GPL
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "SDL.h"
#include "SDL_mmap.h"
#include "charactron.h"
#include "bsicon.h"
#include "utf8.h"

#define REALLOCN(PTR,TYPE,N) ((TYPE *)realloc(PTR,sizeof(TYPE)*(N)))

int render_page(charactron *c, const SDL_mspan *m, 
                mempos_t *start, mempos_t *end);
mempos_t *parse_mmap_text(charactron *c, const SDL_mspan *m);

static const char *filename=NULL;

void setcaptionf(const char *cmdstr, ...)
{
  char buf[1024];
  va_list ap;

  va_start(ap,cmdstr);
    vsprintf(buf,cmdstr,ap);
  va_end(ap);

  SDL_WM_SetCaption(buf,"Reader");
}

void browse_pages(charactron *c, const mempos_t *pages, const SDL_mspan *m)
{
  SDL_Event event;
  int page=1,running=1,redraw=1;

  while(running)
  {
    if(redraw)
    {
      mempos_t p=pages[page];
      render_page(c,m,&p,NULL);
      setcaptionf("%s [%d/%ld]",filename,page,pages[0]);
//      scale_page(c);
      charactron_scale(c);
      SDL_Flip(SDL_GetVideoSurface());
      redraw=0;
    }

    SDL_WaitEvent(&event);
    switch(event.type)
    {
    Uint8 mod;

    case SDL_KEYDOWN:
      mod=SDL_GetModState();
      redraw=1;

      switch(event.key.keysym.sym)
      {
      case SDLK_RIGHT:
      case SDLK_PAGEDOWN:
      case SDLK_SPACE:
        if(mod & KMOD_CTRL)
          page+=10;
        else
          page++;

        if(page > pages[0]) page=pages[0];
        break;

      case SDLK_LEFT:
      case SDLK_PAGEUP:
      case SDLK_BACKSPACE:
        if(mod & KMOD_CTRL)
          page-=10;
        else
          page--;

        if(page < 1) page=1;
        break;

      case SDLK_HOME:
        page=1;
        break;

      case SDLK_END:
        page=pages[0];
        break;

      case SDLK_ESCAPE:
        running=0;
        break;

      default:
        break;
      }
      break;
    case SDL_QUIT:
      running=0;
      break;

    default:
      break;
    }
  }
}

#ifdef _WIN32_WCE
#define DEFAULT_W	240
#define DEFAULT_H	(320-(8*7))
#else
#define DEFAULT_W	320
#define DEFAULT_H	240
#endif

static const SDL_Color black={  0,  0,  0,  0};
static const SDL_Color white={255,255,255,  0};

int scan_color(const char *cstr, SDL_Color *out);

int main(int argc, char *argv[])
{
  int n;
  charactron *c=NULL;

  if(SDL_Init(SDL_INIT_VIDEO)<0)
  {
    fprintf(stderr,"Couldn't init SDL: %s\n",SDL_GetError());
    return(1);
  }

  SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

  BS_SetIcon();

  if(SDL_SetVideoMode(DEFAULT_W, DEFAULT_H, 16, SDL_ANYFORMAT)==NULL)
  {
    fprintf(stderr,"Couldn't set video mode: %s\n",SDL_GetError());
    goto MAIN_FAIL;
  }

//  c=charactron_load(get_font_5x9(),NULL,1);
  c=charactron_load(unifont_load(), NULL, 1);
  if(c==NULL)
  {
    fprintf(stderr,"Couldn't open charactron with default font\n");
    goto MAIN_FAIL;
  }

  for(n=1; n<argc; n++)
  {
    mempos_t *pages=NULL;
    mmap_t len;
    SDL_mspan *mem=NULL;

    struct SDL_mmap *mm;

    if((argc-1) > n)
    {
      SDL_Color cl;
      if(strcmp(argv[n],"-fg")==0)
      {
        n++;
        if(scan_color(argv[n],&cl)<0)
          fprintf(stderr,"Invalid color \"%s\"\n",argv[n]);

        charactron_set_color(c,CHARACTRON_FG,cl);
        continue;
      }
      else if(strcmp(argv[n],"-bg")==0)
      {
        n++;
        if(scan_color(argv[n],&cl)<0)
          fprintf(stderr,"Invalid color \"%s\"\n",argv[n]);

        charactron_set_color(c,CHARACTRON_BG,cl);
        continue;
      }
    }

    mm=SDL_mmap_open(argv[n], SDL_MMAP_RD, SDL_MADVISE_SEQUENTIAL);
    if(mm==NULL)
    {
      fprintf(stderr,"Couldn't mmap %s (%p): %s\n",argv[n],mm,SDL_GetError());
      continue;
    }

    if(SDL_mmap_len(mm,&len)<0)
    {
      fprintf(stderr,"Couldn't read length of file\n");
      goto MAIN_NEXT;
    }

    if(SDL_mmap_maxchunk() < len)
    {
      fprintf(stderr,"Memory model is too small to map this file");
      goto MAIN_NEXT;
    }

    mem=SDL_mmap(mm,(mempos_t)len,SDL_MMAP_RD,0);
    if(mem==NULL)
    {
      fprintf(stderr,"Couldn't map into memory\n");
      goto MAIN_NEXT;
    }

    pages=parse_mmap_text(c,mem);
    if(pages == NULL)
    {
      fprintf(stderr,"Couldn't pageify text!\n");
      goto MAIN_NEXT;
    }

    fprintf(stderr,"Book has %ld pages\n",pages[0]);

	{
      const char *next;
      filename=argv[n];
	  next=strpbrk(filename,"/\\");
      while(next != NULL)
	  {
        filename=next+1;
        next=strpbrk(filename,"/\\");
	  }
	}

    browse_pages(c,pages,mem);

MAIN_NEXT:
    if(pages != NULL)	free(pages);
    if(mem != NULL)	SDL_munmap(mem);
    if(mm != NULL)	SDL_mmap_close(mm);
  }

  charactron_free(c);
  SDL_Quit();
  return(0);

MAIN_FAIL:
  if(c != NULL) charactron_free(c);
  SDL_Quit();
  return(1);
}

/*void print_block(const Uint8 *dat, mempos_t start, mempos_t end)
{
	while(start < end)
	{
		if(dat[start] != '\r')
			fprintf(stderr, "%c", dat[start]);
		start++;
	}
}*/

typedef struct mem_word
{
	mempos_t start, end, glyphs;
} mem_word;

int render_splay(charactron *c, const Uint8 *mem, mempos_t *start_in,
	mempos_t end)
{
	mempos_t start=(*start_in);
	int n=0, g;

	while((g=utf8_getchar(mem, &start, end))>=0)
	{	// Ignore control chars
		if(g < ' ')	continue;

		if(charactron_draw_letter(c, g)<0)
		{
			if(charactron_linefeed(c)<0)
				return(-1);

			if(charactron_draw_letter(c, g)<0)
				return(-1);
		}

		(*start_in)=start;
	}
	
	return(n);
}

int render_page(charactron *c, const SDL_mspan *m,
	mempos_t *start, mempos_t *end)
{
	int started=0;
	int g;
	unifont *u=(c->rawfont);
	mempos_t max, t;
	mem_word word={*start, *start, 0};

	if(end == NULL) max=m->len;
	else            max=(*end);

	charactron_formfeed(c);

	t=word.end;

	while((g=utf8_getchar(m->mem, &t, max))>=0)
	{
		int r;

		switch(g)
		{
		case '\f':
			if(! started)	// Blank pages are stupid
			{
				word.start=word.end=t;
				continue;
			}
			/* FALL THROUGH */
		case '\n':
		case ' ':
			// Word past end of line?  Newline, then try again
			if(c->xfree < word.glyphs)
			if(charactron_linefeed(c) < 0)
				goto RENDER_PAGE_END;

			started+=charactron_draw_data(c, m->mem, 
				&word.start, word.end);

			// Eat the last character
			utf8_getchar(m->mem, &word.start, max);
			word.end=word.start;

			switch(g)
			{
			case '\f':
				goto RENDER_PAGE_END;

			case ' ':
				if((c->xfree) < 2)
				{
					if(charactron_linefeed(c) < 0)
						goto RENDER_PAGE_END;
				}
				else
					CHARACTRON_SPACE(c);
				break;

			case '\n':
				if(started)
				if(charactron_linefeed(c) < 0)
					goto RENDER_PAGE_END;
				break;
			default:
				break;
			}

			word.glyphs=0;
			break;

		default:
			if(UNIFONT_WIDE(u, g))
				word.glyphs++;

			word.glyphs++;
			/* FALL THROUGH */
		case '\t':	// Do not print these chars
		case '\r':
			word.end=t;
			break;
		}

		if(word.glyphs > c->maxx)
		{
//			started+=render_splay(c, m->mem, &word.start, word.end);
			r=render_splay(c, m->mem, &word.start, word.end);
			if(r < 0)
				goto RENDER_PAGE_END;

			started+=r;
			word.start=word.end;
			word.glyphs=0;
		}
	}


	if(c->maxx < word.glyphs)
	{
		started+=render_splay(c, m->mem, &word.start, word.end);
		word.start=word.end;
		word.glyphs=0;
		goto RENDER_PAGE_END;
	}

	// Word past end of line?  Newline, then try again
	if(c->xfree < word.glyphs)
		if(charactron_linefeed(c) < 0)
			goto RENDER_PAGE_END;

	started+=charactron_draw_data(c, m->mem, &word.start, word.end);

RENDER_PAGE_END:
	if(end != NULL)
	{
//		if(word.start < 1000)
		(*end)=word.start;
//		else
//			(*end)=max;
	}

	return(0);
}

mempos_t *parse_mmap_text(charactron *c, const SDL_mspan *m)
{
  Uint32 oldflag=c->flag;
  mempos_t *pages=NULL;
  int numpages=0;
  mempos_t pos=(mempos_t)0;

  CHARACTRON_DISABLE(c);

  while(pos < m->len)
  {
    mempos_t npos=pos,nend=m->len;

//    static int loops=0;

//    fprintf(stderr,"%ld/%ld\n",pos,m->len);

//    loops++;
//    if(loops>2)
//      abort();

//    fprintf(stderr,">>parse_mmap_text: npos=%ld, nend=%ld\n",npos,nend);
    render_page(c,m,&npos,&nend);
//    fprintf(stderr,"<<parse_mmap_text: npos=%ld, nend=%ld\n",npos,nend);

    if(pos > nend)// || (nend == 0L))
      break;
    pos=nend;

    pages=REALLOCN(pages,mempos_t,((++numpages) + 1));
    pages[numpages]=npos;
  }

  pages[0]=numpages;
  c->flag=oldflag;
  return(pages);
}

int scan_color(const char *cstr, SDL_Color *out)
{
  if(sscanf(cstr,"%02hhx%02hhx%02hhx\n",
    &(out->r), &(out->g), &(out->b))<3)
      return(-1);

  out->unused=0;
  return(0);
}
